      subroutine vumat(
C Read only (unmodifiable)variables -
     1  nblock, ndir, nshr, nstatev, nfieldv, nprops, lanneal,
     2  stepTime, totalTime, dt, cmname, coordMp, charLength,
     3  props, density, strainInc, relSpinInc,
     4  tempOld, stretchOld, defgradOld, fieldOld,
     5  stressOld, stateOld, enerInternOld, enerInelasOld,
     6  tempNew, stretchNew, defgradNew, fieldNew,
C Write only (modifiable) variables -
     7  stressNew, stateNew, enerInternNew, enerInelasNew )
C
      include 'VABA_PARAM.INC'
C
      dimension props(nprops), density(nblock), coordMp(nblock,*),
     1  charLength(nblock), strainInc(nblock,ndir+nshr),
     2  relSpinInc(nblock,nshr), tempOld(nblock),
     3  stretchOld(nblock,ndir+nshr),
     4  defgradOld(nblock,ndir+nshr+nshr),
     5  fieldOld(nblock,nfieldv), stressOld(nblock,ndir+nshr),
     6  stateOld(nblock,nstatev), enerInternOld(nblock),
     7  enerInelasOld(nblock), tempNew(nblock),
     8  stretchNew(nblock,ndir+nshr),
     8  defgradNew(nblock,ndir+nshr+nshr),
     9  fieldNew(nblock,nfieldv),
     1  stressNew(nblock,ndir+nshr), stateNew(nblock,nstatev),
     2  enerInternNew(nblock), enerInelasNew(nblock)
      character*80 cmname


* -------------------------------------------------------------------------
* Explicit Arruda-Boyce model for Abaqus for 2D(plane strain)/Axisymmetry
*                       Vegard Tommernes
* -------------------------------------------------------------------------
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                         C  
C     Some explanation:                                                   C
C     Variables                                                           C
C           Deformation gradient          [F]                             C
C           Stress tensor                 [T]                             C
C           strain softening parameter    [s]                             C
C           Stretch Tensor                [U]                             C
C           Rotation Tensor               [R]                             C
C     Type of variables (subscript):                                      C
C           D     means derivative with respect to time                   C
C           e     means elastic                                           C
C           i     means inelastic                                         C
C           n     means time = t      (start of increment)                C
C           nn    means time = t+dt    (end of increment)                 C
C           _ave  means time = t+dt/2  (middle of increment)              C
C     Examples:                                                           C
C       Fin:      Inelastic deformation gradient at start of increment    C
C       Tnn:      Stress at end of increment                              C
C       s_ave:    Strain softening parameter at middle of increment       C
C                                                                         C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC     

C     DEFINING VARIABLES      
c     Arruda Boyce Theory:       
c     Variables taken from abaqus and statevariables:
      Double precision,Dimension(3,3) :: Fn,Fin,U,Fnn
      Double precision                :: sn       
c     Variables given to abaqus and statevariables:       
      Double precision,Dimension(3,3) :: Finn,Tnn
      Double precision                :: snn
c     Variables used in the calculation 
      Double Precision,dimension(3,3) :: Fnn2,Fen2,Fin2,Finn2,Fenn2,Fn2
      Double precision,Dimension(3,3) :: R,Fen,Fenn,F_ave,Fe_ave
      Double precision,Dimension(3,3) :: Fi_ave,DFi_ave,DFin,DFin2
      Double precision                :: s_ave,Ds_ave
      Double Precision                :: Dsn,sn2,snn2,Dsn2
      integer                         :: i,j 
c     The material constants
      double precision  ::  E,nu,gammadot0,A,h,sfrac,Cr,N,G
      double precision  ::  boltz,abstemp,sss,flowconst,s0,Lambda
c     lumping material constants together in a vector:      
      Double precision,Dimension(8) :: mprop       
      
c     Iteration:          
c     variables defined for the iteration process:    
      Double precision                ::  kk,b,c
      integer                         ::  MaxIter
c     variables used in the iteration process  
      Double Precision,dimension(3,3) :: FFE,Tb,Tvp
      Double Precision                :: eps,temp1,temp2,MinItStep,diff
      Double precision                :: dt_new,dt_sugg,dt_temp
      integer                         :: k,iter,k2,iter2

C**************************************************************************
C********* Defining material and iteration parameter **********************
C**************************************************************************
           
C     Defining all material material properties:            
      E             = 2300                         !   MPa   
      nu            = 0.33                         !                  
      gammadot0     = 2.0e15                       !  
      A             = 3.31e-18                     !   mm^3
      h             = 500                          !   MPa
      sfrac         = 0.78                         !      
      Cr            = 18.0                         !   MPa
      N             = 2.78                         !   
        
      boltz         = 1.3806503e-20                !     
      abstemp       = 273+22.5                     !   K
       
      G             = E/(2*(1+nu))                 !   MPa 
      s0            = 0.11*G                       !   MPa
      sss           = sfrac * s0                   !   MPa
      flowconst     = A/(abstemp*boltz)            !  (ms^2)kg
      Lambda        = (E*nu)/((1+nu)*(1-2*nu))     !   MPa
      
C     Lumping relevant material properties in a vector mprop
      mprop(1)          =   G
      mprop(2)          =   Lambda
      mprop(3)          =   Cr   
      mprop(4)          =   N
      mprop(5)          =   gammadot0
      mprop(6)          =   h
      mprop(7)          =   sss
      mprop(8)          =   flowconst
      
C     Iteration     
      kk        = 0.1    !smaller kk -> more iterations -> more accurate
      b         = 0.95   !saftey factor
      c         = 0.5    !safety factor
      MaxIter   = 10000  !Max iterations
      MinItStep = 1e-10  !avoid deviding by zero
      diff      = 1e-10  !size of increment     

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

c     Abaqus vumat: do for all elements
      do 100 km = 1,nblock           
      
C**************************************************************************
C*********************** Initializing variables ***************************
C**************************************************************************

      sn       = 0
      snn      = 0
      s_ave    = 0
      Ds_ave   = 0
      Dsn      = 0
      sn2      = 0
      snn2     = 0
      Dsn2     = 0
      eps      = 0
      temp1    = 0
      temp2    = 0
      dt_new   = 0
      dt_sugg  = 0
      dt_temp  = 0
      
      do i = 1,3
        do j = 1,3
          Fn(i,j)       = 0
          F_ave(i,j)    = 0
          Fnn(i,j)      = 0
          U(i,j)        = 0
          Tnn(i,j)      = 0          
          Fin(i,j)      = 0
          Fi_ave(i,j)   = 0
          Finn(i,j)     = 0
          Fen(i,j)      = 0 
          Fe_ave(i,j)   = 0         
          Fenn(i,j)     = 0
          DFin(i,j)     = 0
          DFi_ave(i,j)  = 0
          FFE           = 0
          Fn2(i,j)      = 0
          Fnn2(i,j)     = 0
          Fin2(i,j)     = 0
          Finn2(i,j)    = 0
          Fen2(i,j)     = 0
          Fenn2(i,j)    = 0
          DFin2(i,j)    = 0
        end do
      end do  
      
C**************************************************************************
C*************** Getting variables from abaqus ****************************
C**************************************************************************      
      
c     get Fn from defgradold
      Fn(1,1) = defgradold(km,1)
      Fn(2,2) = defgradold(km,2)
      Fn(3,3) = defgradold(km,3)
      Fn(1,2) = defgradold(km,4)
      Fn(2,1) = defgradold(km,5)
      
c     get Fnn from defgradnew   
      Fnn(1,1) = defgradnew(km,1)
      Fnn(2,2) = defgradnew(km,2)
      Fnn(3,3) = defgradnew(km,3)
      Fnn(1,2) = defgradnew(km,4)
      Fnn(2,1) = defgradnew(km,5)
      
c     get the stretch tensor U from stretchnew
c     Not needed until after the iteration.      
      U(1,1) = stretchnew(km,1)
      U(2,2) = stretchnew(km,2)
      U(3,3) = stretchnew(km,3)
      U(1,2) = stretchnew(km,4)
      U(2,1) = U(1,2)         
           
c     get Fin from statev, Statev = Fin - I
      do i = 1,3
        Fin(i,i) = 1.0 + StateOld(km,i)
      end do
      Fin(1,2) = StateOld(km,4)
      Fin(2,1) = StateOld(km,5)

c     Get Strain softening parameter from statev: Statev = snn - 1       
      sn = s0 + StateOld(km,6)

c     calculate the inelastic deformation gradient at start of increment
      call MatProdInv(Fn,Fin,Fen)
      
c     Get the rotation tensor R      
      call MatProdInv(Fnn,U,R)            

C**************************************************************************
C*************** Main Calculations ****************************************
C**************************************************************************

c     Check if this is the first increment
      if (steptime .eq. 0) then
        Finn = Fin
        snn  = sn
        call MatProdInv(Fnn,Finn,Fenn)
        call ElasticStress(10*mprop,Fenn,Tnn)
      goto 12
      end if
     
c     We are not in the first increment -> Viscoplastic response  
       call FlowRule(mprop,Fen,
     1     Fin,DFin,sn,Dsn,Tb,Tvp)
           Fi_ave = Fin + 0.5*DT*DFin
           s_ave  = sn  + 0.5*DT*Dsn
           F_ave = 0.5*(Fnn + Fn)    
      call MatProdInv(F_ave,Fi_ave,Fe_ave)
      call FlowRule(mprop,Fe_ave,
     1     Fi_ave,DFi_ave,s_ave,Ds_ave,Tb,Tvp)
           Finn = Fin + DT * DFi_ave  !Finn using Midpoint scheeme. 
           snn  = sn  + DT * Ds_ave
            
C     uncomment "goto 10" to accept the calculations from Midpoint
C     and skip the "accuracy check and iteration scheme". 

C         goto 10      
           
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC           
CCCCCCCCCC ACCURACY CHECK AND ITERATION CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
               
           FFE = Fin + dt*DFin      !Finn using Forward Euler(1. order)
      call MatNorm(FFE-Finn,temp1)
      call MatNorm(Finn-Fin,temp2)
      if (temp2 .lt. diff) goto 10
              eps = temp1/temp2  
          if (eps .ne. eps) then
              print*, " "
              print*, "epsilon = NaN"
              print*, "Finn is not defined correctly"
              print*, "possible cause: Flow Rule blows up"
              pause
          end if    
          if (eps .gt. kk) then    !Error is too big!
              dt_new = b*DT*(kk/eps)
              dt_temp = min(dt_new,c*DT)
              iter = ceiling(DT/dt_temp)
              dt_sugg = DT/iter
c             Same theory but with lower timestep
   20             if (iter .gt. MaxIter) goto 10
                  Fn2  = Fn              
                  Fen2 = Fen
                  Fin2 = Fin
                  sn2  = sn
               do k = 1,iter
                          k2 = k
                          iter2 = iter
                          Fnn2 = Fn + (k2/iter2)*(Fnn-Fn)
                          F_ave = 0.5 * (Fnn2 + Fn2)
                     call FlowRule(mprop,Fen2,
     1                    Fin2,DFin2,sn2,Dsn2,Tb,Tvp)
                          Fi_ave = Fin2 + 0.5*dt_sugg * Dfin2
                          s_ave  = sn2  + 0.5*dt_sugg * Dsn2
                     call MatProdInv(F_ave,Fi_ave,Fe_ave)
                     call FlowRule(mprop,Fe_ave,
     1                    Fi_ave,DFi_ave,s_ave,Ds_ave,Tb,Tvp)
                          Finn2 = Fin2 + dt_sugg * DFi_ave
                          snn2 = sn2 + dt_sugg * Ds_ave  
                     call MatProdInv(Fnn2,Finn2,Fenn2)      
                                   !Checking Error again:
                                   FFE = Fin2 + dt_sugg*DFin2
                              call MatNorm(FFE-Finn2,temp1)
                              call MatNorm(Finn2-Fin2,temp2)
                              if (temp2 .lt. MinItStep) goto 30     
                                  eps = temp1/temp2    
                              if (eps .gt. kk) then          
                                  dt_new = b*dt_sugg*(kk/eps)
                                  dt_temp = min(dt_new,c*dt_sugg)
                                  iter = ceiling(DT/dt_temp)
                                  dt_sugg = DT/iter
                                  goto 20
                              end if
  30                      Fn2 = Fnn2
                          Fen2 = Fenn2
                          Fin2 = Finn2
                          sn2  = snn2
              end do 
                 Fenn = Fenn2
                 Finn = Finn2
                 snn  = snn2
          end if   
                 
CCCCCCCCCC END ACCURACY CHECK AND ITERATION CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC           
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC          
          
      !calculate Fenn using Fnn and Finn, and calculate the stress Tnn    
  10  call MatProdInv(Fnn,Finn,Fenn)
      call ElasticStress(mprop,Fenn,Tnn)
      
c     Rotate the stress tensor:      
12    call TMatProd3(R,Tnn,R,Tnn)
  
C**************************************************************************
C******** Main Calculations Finished -> Storing variables *****************
C**************************************************************************  

c     Updating the stress  
      stressnew(km,1) = Tnn(1,1)
      stressnew(km,2) = Tnn(2,2)
      stressnew(km,3) = Tnn(3,3)
      stressnew(km,4) = Tnn(1,2)    
      
c     store statevariables
      do i = 1,3
        Statenew(km,i) = Finn(i,i) - 1.0
      end do
      Statenew(km,4) = Finn(1,2)
      Statenew(km,5) = Finn(2,1)
      StateNew(km,6) = snn - s0

  100 continue
      

      return
      end
 
 
      
C**************************************************************************
C**************** Arruda Boyce Theory - subroutines ***********************
C**************************************************************************


CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
CC            FLOWRULE AND VISCOPLASTIC DAMPENING                          CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      Subroutine FlowRule(mprop,Fe,Fi,DFi,s,sdot,Tb,Tvp)
          implicit none
          Double precision :: Tr_Tvp,tau,gammadotp
          Double precision :: Tr_Dev_Tvp2,s,sdot,temp,const
          Double precision :: gammadot0,h,flowconst,s0,sss
          Double precision,Dimension(3,3)::F,Fe,Fi,DFi,Te
          Double precision,Dimension(3,3)::B,Be,Bi,Tb
          Double precision,Dimension(3,3)::Tvp, Dev_Tvp, dev_Tvp2
          Double precision,Dimension(3,3)::Identity,Nij,D,R
          Double precision,Dimension(8)::mprop  
          integer::k,l         
c     material parameters                    
          gammadot0         =   mprop(5)
          h                 =   mprop(6)
          sss               =   mprop(7)
          flowconst         =   mprop(8)
c     Preparations   
          Tr_Tvp = 0.0
          temp   = 0.0
          tau    = 0.0
          do k = 1,3
            do l = 1,3
              D(k,l)        = 0.0
              Identity(k,l) = 0.0
              Te(k,l)       = 0.0
              Tvp(k,l)      = 0.0
              dev_Tvp(k,l)  = 0.0
              Nij(k,l)      = 0.0
            end do 
          end do     
C     Main calculations          
          call MatIdent(Identity)
          call ElasticStress(mprop,Fe,Te)
          call FeBackStressFeT(mprop,Fe,Fi,Tb)
               Tvp = Te - Tb  
          call MatTrace(Tvp,Tr_Tvp)         
               dev_Tvp = Tvp - (Tr_Tvp/3)*Identity
               do k = 1,3
                do l = 1,3
                   temp = temp + dev_tvp(k,l)**2.0 
                end do 
               end do
               tau = sqrt(0.5)*sqrt(temp)   
          if (tau .gt. 0) then      !We have plastic deformation
                Nij=(sqrt(0.5)/tau)*dev_Tvp 
          end if 
               gammadotp=gammadot0*
     1         exp(-(flowconst*s)*(1-((tau/s)**(5.0/6.0))))
               D = gammadotp*Nij
          call MatProd(D,Fi,DFi)
               sdot = h*(1-(s/sss))*gammadotp
          return 
      end subroutine FlowRule


CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
CC              CALCULATE THE BACKSTRESS                                   CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
      Subroutine FeBackStressFeT(mprop,Fe,Fi,Tb)
      implicit none 
          Double precision::Tr_Bi, Lambdapch, Cr, N, temp, IL,I,IV   
          Double precision,Dimension(3,3)::Fe,Fi,Fe_Tb_FeT,Bi
          Double precision,Dimension(8)::mprop
          Double precision,dimension(3,3)::Q,BB,Tb,C,C_hat,U,V,R
          Double precision,dimension(3)::EigV,EigC
          Double precision::third,detFe
          integer::k,l
c     material parameters          
          Cr           = mprop(3)
          N            = mprop(4)
c     Preparations
          Lambdapch = 0.0
          temp      = 0.0
          IL        = 0.0 
          do k = 1,3
                EigV(k) = 0.0
                do l = 1,3
                    BB(k,l)        = 0.0
                    Bi(k,l)        = 0.0
                    Q(k,l)         = 0.0
                    Fe_Tb_FeT(k,l) = 0.0
                end do 
          end do
C     Main Calculations          
          call MatDet(Fe,detFe)         
          call MatIdent(C_hat)
          call TMatProd(Fi,Fi,C)
          call DSYEVJ3(C,Q,eigC)
          do k = 1,3
            c_hat(k,k) = sqrt(eigC(k))
          end do 
          call TMatProd3(Q,c_hat,Q,U)
          call MatProdInv(Fi,U,R)
          call MatProdInv(Fi,R,V)
          Q = 0.0
          call DSYEVJ3(C,Q,EigV)
               IV = EigV(1) + EigV(2) + EigV(3)  
               Lambdapch=sqrt(IV/3.0)
               temp=Lambdapch/sqrt(N)
               call InvLang(Temp,IL)
            do k=1,3
         BB(k,k)=(Cr/3.0)*sqrt(N)*IL*(EigV(k)-IV/3.0)/Lambdapch
            end do 
          call MatProdT3(Fe,BB,Fe,Fe_Tb_FeT)
          Tb = (1.0/detFe)*Fe_Tb_FeT
          return 
      end subroutine FeBackStressFeT                    
             
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
CC          CALCULATE THE ELASTIC STRESS RESPONSE                          CC
CC      This subroutine uses the material properties and the elastic       CC
CC      deformation gradient to calculate the elastic stress               CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
      Subroutine ElasticStress(mprop,Fe,Te)
      include 'VABA_PARAM.INC'
          Double precision :: DetFe,G,Lambda,Tr_lnsqrtBe       
          Double precision,Dimension(3,3)::Te,Fe,Be,lnsqrtBe
          Double precision,Dimension(3,3)::temp
          Double precision,Dimension(8)::mprop
          integer :: i,j           
c     Material properties:
          G       = mprop(1)
          Lambda  = mprop(2)      
c     preparations: 
          DetFe = 0.0              
          do i = 1,3
            do j = 1,3
                Te(i,j)       = 0.0
                temp(i,j)     = 0.0
                Be(i,j)       = 0.0
                lnsqrtBe(i,j) = 0.0
            end do
          end do           
c     Main calculations                       
          call MatDet(Fe,DetFe)
          call MatProdT(Fe,Fe,Be)
          call TMatProd(Fe,Fe)
          !Using Pade approximation or analytical solution to find: henky strain = ln(sqrt(Be))
          !call PadeApprox(Be,lnsqrtBe)
          call PadeExact(Be,lnsqrtBe)
          tr_lnsqrtBe = lnsqrtBe(1,1)+lnsqrtBe(2,2)+lnsqrtBe(3,3)
          do i = 1,3
            do j = 1,3
                Te(i,j) = 2*G*lnsqrtBe(i,j)
                if (i .eq. j) then
                    Te(i,j) = Te(i,j) + Lambda*tr_lnsqrtBe
                end if
            end do 
          end do  
          temp = Te
          Te = (1.0/DetFe)*Temp 
          return 
      end subroutine ElasticStress  

C**************************************************************************
C********************* Numerical approximations ***************************
C**************************************************************************

C     APPROXIMATE (ln(sqrt(Mat)))
c       This subroutine uses Pade approximation
c       to find (ln(sqrt(Mat))) of Mat      
      subroutine PadeApprox(Mat,lnsqrtMat)
      include 'VABA_PARAM.INC'
      double precision,dimension(3,3)::I,lnsqrtMat,Mat
      double precision,dimension(3,3)::a,b,c,c_inv,Mat2,y,x
      call MatIdent(I)
      a=Mat+I
      b=Mat-I
      call MatProd(Mat,Mat,Mat2)
      c = 2 * ( Mat2 + 4 * Mat + I)
      call MatInv(c,c_inv)
      call MatProd(a,b,x)
      call MatProd(x,c_inv,y)
      lnsqrtMat=3*y
      end subroutine PadeApprox
      
C     APPROXIMATE (ln(sqrt(Mat)))
c       This subroutine uses Pade approximation
c       to find (ln(sqrt(Mat))) of Mat      
      subroutine PadeExact(Mat,lnsqrtMat)
      include 'VABA_PARAM.INC'
      double precision,dimension(3,3)::Mat,lnsqrtMat,Q,temp
      double precision,dimension(3)::W,Q_inv
      call DSYEVJ3(Mat,Q,W)
      call MatIdent(temp)
      temp(1,1) = 0.5*log(W(1))
      temp(2,2) = 0.5*log(W(2))
      temp(3,3) = 0.5*log(W(3))
      Call MatProdInv3(Q,temp,Q,lnsqrtMat)
      end subroutine PadeExact            
      
C     This gives an approximation to the inverse langevin function
C     The function takes in: a, and gives the result as b
C     within 5% accuracy 
      Subroutine InvLang(x,L)
      include 'VABA_PARAM.INC'
        double precision :: x,L,a,b,c
        a = 2.99248834685337 
        b = -1.14365108190676
        c = -1.000
        L = x*((a+b*x*x)/(1+c*x*x))        
        return
      end Subroutine InvLang 
      
     
C**************************************************************************
C****************** SUBROUTINES - Matrix operations ***********************
C**************************************************************************            

C     calculate the product of a*b*trans(c):       
      Subroutine MatProdInv3(a,b,c,d)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,d,x,y
      call MatProd(a,b,x)
      call MatInv(c,y)
      call MatProd(x,y,d)
      return 
      end subroutine MatProdInv3

C     calculate the product of trans(a)*b*c:       
      Subroutine TMatProd3(a,b,c,d)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,d,x,y
      call MatTrans(a,y)
      call MatProd(y,b,x)
      call MatProd(x,c,d)
      return 
      end subroutine TMatProd3
      
C     calculate the product of a*b*trans(c):       
      Subroutine MatProdT3(a,b,c,d)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,d,x,y
      call MatProd(a,b,x)
      call MatTrans(c,y)
      call MatProd(x,y,d)
      return 
      end subroutine MatProdT3      
      
C     calculate the product of a matrix a and inv(b)     
      Subroutine MatProdInv(a,b,c)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,x
      call MatInv(b,x)
      call MatProd(a,x,c)
      return 
      end subroutine MatProdInv         

C     calculate the product of a matrix a and trans(b)     
      Subroutine TMatProd(a,b,c)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,x
      call MatTrans(a,x)
      call MatProd(x,b,c)
      return 
      end subroutine TMatProd 

C     calculate the product of a matrix a and trans(b)     
      Subroutine MatProdT(a,b,c)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,x
      call MatTrans(b,x)
      call MatProd(a,x,c)
      return 
      end subroutine MatProdT 

C     Calculating the inverse of a 3x3 matrix
      Subroutine MatInv(a,b)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3) :: a,b,x,z 
      Double precision :: y
      call MatDet(a,y)
      call MatTrans(a,x)
          z(1,1) =  +(x(2,2)*x(3,3) - x(2,3)*x(3,2))
          z(1,2) =  -(x(2,1)*x(3,3) - x(2,3)*x(3,1))
          z(1,3) =  +(x(2,1)*x(3,2) - x(2,2)*x(3,1))
          z(2,1) =  -(x(1,2)*x(3,3) - x(1,3)*x(3,2))
          z(2,2) =  +(x(1,1)*x(3,3) - x(1,3)*x(3,1))
          z(2,3) =  -(x(1,1)*x(3,2) - x(1,2)*x(3,1))
          z(3,1) =  +(x(1,2)*x(2,3) - x(1,3)*x(2,2))
          z(3,2) =  -(x(1,1)*x(2,3) - x(1,3)*x(2,1))
          z(3,3) =  +(x(1,1)*x(2,2) - x(1,2)*x(2,1))
      b = (1/y) * z    
      return
      end subroutine MatInv  
      
C     calculating transpose of matrix
      Subroutine MatTrans(a,b)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3) :: a,b
      integer j,k
      do j=1,3
        do k=1,3
            b(j,k) = a(k,j)
        end do 
      end do
      return
      end Subroutine MatTrans      
      
C     Calculate the determinant of a matrix
      Subroutine MatDet(a,det)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a 
      Double precision :: det
      det = a(1,1)*a(2,2)*a(3,3)-a(1,1)*a(2,3)*a(3,2)-
     1 a(1,2)*a(2,1)*a(3,3)+ a(1,2)*a(2,3)*a(3,1)+
     2 a(1,3)*a(2,1)*a(3,2)- a(1,3)*a(2,2)*a(3,1)
      return
      end Subroutine MatDet
      
C     Calculating product of two 3x3 matrices:
      Subroutine MatProd(a,b,c)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c
          c(1,1) = a(1,1)*b(1,1) + a(1,2)*b(2,1) + a(1,3)*b(3,1)
          c(1,2) = a(1,1)*b(1,2) + a(1,2)*b(2,2) + a(1,3)*b(3,2)
          c(1,3) = a(1,1)*b(1,3) + a(1,2)*b(2,3) + a(1,3)*b(3,3)
          c(2,1) = a(2,1)*b(1,1) + a(2,2)*b(2,1) + a(2,3)*b(3,1)
          c(2,2) = a(2,1)*b(1,2) + a(2,2)*b(2,2) + a(2,3)*b(3,2)
          c(2,3) = a(2,1)*b(1,3) + a(2,2)*b(2,3) + a(2,3)*b(3,3)
          c(3,1) = a(3,1)*b(1,1) + a(3,2)*b(2,1) + a(3,3)*b(3,1)
          c(3,2) = a(3,1)*b(1,2) + a(3,2)*b(2,2) + a(3,3)*b(3,2)
          c(3,3) = a(3,1)*b(1,3) + a(3,2)*b(2,3) + a(3,3)*b(3,3)
      return
      end Subroutine MatProd
      
c     This makes a 3x3 identity matrix      
      Subroutine MatIdent(I)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::I
      integer::k,l
      do k = 1,3
         do l = 1,3
            if (k .eq. l) then
                I(k,l) = 1.0
            else
                I(k,l) = 0.0
            end if
         end do 
      end do 
      return
      end Subroutine MatIdent
      
c     This subroutine calculates the trace of a matrix
      Subroutine MatTrace(Mat,Trace)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)   ::  Mat
      Double precision                  ::  Trace
      Trace = Mat(1,1)+Mat(2,2)+Mat(3,3)
      return
      end Subroutine MatTrace
      
c     This subroutine calculates the norm of a matrix
      Subroutine MatNorm(Mat,Norm)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)   ::  Mat
      Double precision                  ::  temp,Norm
      integer::i,j
      temp = 0
      do i = 1,3
        do j = 1,3
            temp = temp + Mat(i,j)**2
        end do 
      end do
      Norm = sqrt(temp)
      return
      end Subroutine MatNorm      
            
      
C**************************************************************************
C****************  EIGENVALUES and EIGENMATRIX ****************************
C**************************************************************************            

* ----------------------------------------------------------------------------
* Numerical diagonalization of 3x3 matrcies
* Copyright (C) 2006  Joachim Kopp
* ----------------------------------------------------------------------------
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
* ----------------------------------------------------------------------------


* ----------------------------------------------------------------------------
      SUBROUTINE DSYEVJ3(A, Q, W)
* ----------------------------------------------------------------------------
* Calculates the eigenvalues and normalized eigenvectors of a symmetric 3x3
* matrix A using the Jacobi algorithm.
* The upper triangular part of A is destroyed during the calculation,
* the diagonal elements are read but not destroyed, and the lower
* triangular elements are not referenced at all.
* ----------------------------------------------------------------------------
* Parameters:
*   A: The symmetric input matrix
*   Q: Storage buffer for eigenvectors
*   W: Storage buffer for eigenvalues
* ----------------------------------------------------------------------------
*     .. Arguments ..
      DOUBLE PRECISION A(3,3)
      DOUBLE PRECISION Q(3,3)
      DOUBLE PRECISION W(3)

*     .. Parameters ..
      INTEGER          N
      PARAMETER        ( N = 3 )
    
*     .. Local Variables ..
      DOUBLE PRECISION SD, SO
      DOUBLE PRECISION S, C, T
      DOUBLE PRECISION G, H, Z, THETA
      DOUBLE PRECISION THRESH
      INTEGER          I, X, Y, R

*     Initialize Q to the identitity matrix
*     --- This loop can be omitted if only the eigenvalues are desired ---
      DO 10 X = 1, N
        Q(X,X) = 1.0D0
        DO 11, Y = 1, X-1
          Q(X, Y) = 0.0D0
          Q(Y, X) = 0.0D0
   11   CONTINUE
   10 CONTINUE

*     Initialize W to diag(A)
      DO 20 X = 1, N
        W(X) = A(X, X)
   20 CONTINUE

*     Calculate SQR(tr(A))  
      SD = 0.0D0
      DO 30 X = 1, N
        SD = SD + ABS(W(X))
   30 CONTINUE
      SD = SD**2
 
*     Main iteration loop
      DO 40 I = 1, 50
*       Test for convergence
        SO = 0.0D0
        DO 50 X = 1, N
          DO 51 Y = X+1, N
            SO = SO + ABS(A(X, Y))
   51     CONTINUE
   50   CONTINUE
        IF (SO .EQ. 0.0D0) THEN
          RETURN
        END IF

        IF (I .LT. 4) THEN
          THRESH = 0.2D0 * SO / N**2
        ELSE
          THRESH = 0.0D0
        END IF

*       Do sweep
        DO 60 X = 1, N
          DO 61 Y = X+1, N
            G = 100.0D0 * ( ABS(A(X, Y)) )
            IF ( I .GT. 4 .AND. ABS(W(X)) + G .EQ. ABS(W(X))
     $                    .AND. ABS(W(Y)) + G .EQ. ABS(W(Y)) ) THEN
              A(X, Y) = 0.0D0
            ELSE IF (ABS(A(X, Y)) .GT. THRESH) THEN
*             Calculate Jacobi transformation
              H = W(Y) - W(X)
              IF ( ABS(H) + G .EQ. ABS(H) ) THEN
                T = A(X, Y) / H
              ELSE
                THETA = 0.5D0 * H / A(X, Y)
                IF (THETA .LT. 0.0D0) THEN
                  T = -1.0D0 / (SQRT(1.0D0 + THETA**2) - THETA)
                ELSE
                  T = 1.0D0 / (SQRT(1.0D0 + THETA**2) + THETA)
                END IF
              END IF

              C = 1.0D0 / SQRT( 1.0D0 + T**2 )
              S = T * C
              Z = T * A(X, Y)
              
*             Apply Jacobi transformation
              A(X, Y) = 0.0D0
              W(X)    = W(X) - Z
              W(Y)    = W(Y) + Z
              DO 70 R = 1, X-1
                T       = A(R, X)
                A(R, X) = C * T - S * A(R, Y)
                A(R, Y) = S * T + C * A(R, Y)
   70         CONTINUE
              DO 80, R = X+1, Y-1
                T       = A(X, R)
                A(X, R) = C * T - S * A(R, Y)
                A(R, Y) = S * T + C * A(R, Y)
   80         CONTINUE
              DO 90, R = Y+1, N
                T       = A(X, R)
                A(X, R) = C * T - S * A(Y, R)
                A(Y, R) = S * T + C * A(Y, R)
   90         CONTINUE

*             Update eigenvectors
*             --- This loop can be omitted if only the eigenvalues are desired ---
              DO 100, R = 1, N
                T       = Q(R, X)
                Q(R, X) = C * T - S * Q(R, Y)
                Q(R, Y) = S * T + C * Q(R, Y)
  100         CONTINUE
            END IF
   61     CONTINUE
   60   CONTINUE
   40 CONTINUE

      PRINT *, "DSYEVJ3: No convergence."
            
      END SUBROUTINE
* End of subroutine DSYEVJ3