      subroutine vumat(
C Read only (unmodifiable)variables -
     1  nblock, ndir, nshr, nstatev, nfieldv, nprops, lanneal,
     2  stepTime, totalTime, dt, cmname, coordMp, charLength,
     3  props, density, strainInc, relSpinInc,
     4  tempOld, stretchOld, defgradOld, fieldOld,
     5  stressOld, stateOld, enerInternOld, enerInelasOld,
     6  tempNew, stretchNew, defgradNew, fieldNew,
C Write only (modifiable) variables -
     7  stressNew, stateNew, enerInternNew, enerInelasNew )
C
      include 'VABA_PARAM.INC'
C
      dimension props(nprops), density(nblock), coordMp(nblock,*),
     1  charLength(nblock), strainInc(nblock,ndir+nshr),
     2  relSpinInc(nblock,nshr), tempOld(nblock),
     3  stretchOld(nblock,ndir+nshr),
     4  defgradOld(nblock,ndir+nshr+nshr),
     5  fieldOld(nblock,nfieldv), stressOld(nblock,ndir+nshr),
     6  stateOld(nblock,nstatev), enerInternOld(nblock),
     7  enerInelasOld(nblock), tempNew(nblock),
     8  stretchNew(nblock,ndir+nshr),
     8  defgradNew(nblock,ndir+nshr+nshr),
     9  fieldNew(nblock,nfieldv),
     1  stressNew(nblock,ndir+nshr), stateNew(nblock,nstatev),
     2  enerInternNew(nblock), enerInelasNew(nblock)
      character*80 cmname

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C
C     Some explanation:
C     Variables
c           Deformation gradient:         [F]
c           Stress tensor:                [T]
c           strain softening parameter:   [s]
c           Stretch Tensor:               [U]
c           Rotation Tensor               [R]
c     Type of variables
c           D     means derivative with respect to time
c           e     means elastic 
c           i     means inelastic
c           n     means time = t      (start of increment)
c           nn    means time = t+dt    (end of increment)
c           _ave  means time = t+dt/2  (middle of increment)
c     Examples:
c       Fin:      Inelastic deformation gradient at start of increment 
c       Tnn:      Stress at end of increment
c       s_ave:    Strain softening parameter at middle of increment 
C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC     

C     DEFINING VARIABLES      
c     Arruda Boyce Theory:       
c     Variables taken from abaqus and statevariables:
      Double precision,Dimension(3,3) :: Fn,Fin,U,Fnn
      Double precision                :: sn       
c     Variables given to abaqus and statevariables:       
      Double precision,Dimension(3,3) :: Finn,Tnn
      Double precision                :: snn
c     Variables used in the calculation 
      Double Precision,dimension(3,3) :: Fnn2,Fen2,Fin2,Finn2,Fenn2,Fn2
      Double precision,Dimension(3,3) :: R,Fen,Fenn,F_ave,Fe_ave
      Double precision,Dimension(3,3) :: Fi_ave,DFi_ave,DFin,DFin2
      Double precision                :: s_ave,Ds_ave
      Double Precision                :: Dsn,sn2,snn2,Dsn2
      integer                         :: i,j 
c     The material constants
      double precision  ::  E,nu,gammadot0,A,h,sfrac,Cr,N,G
      double precision  ::  boltz,abstemp,sss,flowconst,s0,Lambda
c     lumping material constants together in a vector:      
      Double precision,Dimension(8) :: mprop       
      
c     Iteration:          
c     variables defined for the iteration process:    
      Double precision                ::  kk,b,c
      integer                         ::  MaxIter
c     variables used in the iteration process  
      Double Precision,dimension(3,3) :: FFE
      Double Precision                :: eps,temp1,temp2,MinItStep,diff
      Double precision                :: dt_new,dt_sugg,dt_temp
      integer                         :: k,iter,k2,iter2

C**************************************************************************
C********* Defining material and iteration parameter **********************
C**************************************************************************           
C     Defining all material material properties:            
      E             = 3205                         !   MPa   
      nu            = 0.33                         !                  
      gammadot0     = 2.8e7                       !  
      A             = 1.39e-18                     !   mm^3
      h             = 315                          !   MPa
      sfrac         = 0.87                         !      
      Cr            = 8.0                         !   MPa
      N             = 2.1                         !   
        
      boltz         = 1.3806503e-20                !     
      abstemp       = 273+22.5                     !   K
       
      G             = E/(2*(1+nu))                 !   MPa 
      s0            = 0.11*G                       !   MPa
      sss           = sfrac * s0                   !   MPa
      flowconst     = A/(abstemp*boltz)            !  (ms^2)kg
      Lambda        = (E*nu)/((1+nu)*(1-2*nu))     !   MPa
      
C     Lumping relevant material properties in a vector mprop
      mprop(1)          =   G
      mprop(2)          =   Lambda
      mprop(3)          =   Cr   
      mprop(4)          =   N
      mprop(5)          =   gammadot0
      mprop(6)          =   h
      mprop(7)          =   sss
      mprop(8)          =   flowconst
      
C     Iteration     
      kk        = 0.1    !smaller kk -> more iterations -> more accurate
      b         = 0.95   !saftey factor
      c         = 0.5    !safety factor
      MaxIter   = 10000  !Max iterations
      MinItStep = 1e-10  !avoid deviding by zero
      diff      = 1e-10  !size of increment     

CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

c     Abaqus vumat: do for all elements
      do 100 km = 1,nblock           
      
C**************************************************************************
C*********************** Initializing variables ***************************
C**************************************************************************

      sn       = 0
      snn      = 0
      s_ave    = 0
      Ds_ave   = 0
      Dsn      = 0
      sn2      = 0
      snn2     = 0
      Dsn2     = 0
      eps      = 0
      temp1    = 0
      temp2    = 0
      dt_new   = 0
      dt_sugg  = 0
      dt_temp  = 0
      
      do i = 1,3
        do j = 1,3
          Fn(i,j)       = 0
          F_ave(i,j)    = 0
          Fnn(i,j)      = 0
          U(i,j)        = 0
          Tnn(i,j)      = 0          
          Fin(i,j)      = 0
          Fi_ave(i,j)   = 0
          Finn(i,j)     = 0
          Fen(i,j)      = 0 
          Fe_ave(i,j)   = 0         
          Fenn(i,j)     = 0
          DFin(i,j)     = 0
          DFi_ave(i,j)  = 0
          FFE           = 0
          Fn2(i,j)      = 0
          Fnn2(i,j)     = 0
          Fin2(i,j)     = 0
          Finn2(i,j)    = 0
          Fen2(i,j)     = 0
          Fenn2(i,j)    = 0
          DFin2(i,j)    = 0
        end do
      end do  
      
C**************************************************************************
C*************** Getting variables from abaqus ****************************
C**************************************************************************      
      
c     get Fn from defgradold
      Fn(1,1) = defgradold(km,1)
      Fn(2,2) = defgradold(km,2)
      Fn(3,3) = defgradold(km,3)
      Fn(1,2) = defgradold(km,4)
      Fn(2,1) = defgradold(km,5)
      
c     get Fnn from defgradnew   
      Fnn(1,1) = defgradnew(km,1)
      Fnn(2,2) = defgradnew(km,2)
      Fnn(3,3) = defgradnew(km,3)
      Fnn(1,2) = defgradnew(km,4)
      Fnn(2,1) = defgradnew(km,5)
      
c     get the stretch tensor U from stretchnew
c     Not needed until after the iteration.      
      U(1,1) = stretchnew(km,1)
      U(2,2) = stretchnew(km,2)
      U(3,3) = stretchnew(km,3)
      U(1,2) = stretchnew(km,4)
      U(2,1) = U(1,2)         
           
c     get Fin from statev, Statev = Fin - I
      do i = 1,3
        Fin(i,i) = 1.0 + StateOld(km,i)
      end do
      Fin(1,2) = StateOld(km,4)
      Fin(2,1) = StateOld(km,5)

c     Get Strain softening parameter from statev: Statev = snn - 1       
      sn = s0 + StateOld(km,6)

c     calculate the inelastic deformation gradient at start of increment
      call MatProdInv(Fn,Fin,Fen)
      
c     Get the rotation tensor R      
      call MatProdInv(Fnn,U,R)            

C**************************************************************************
C*************** Main Calculations ****************************************
C**************************************************************************

c     Check if this is the first increment
      if (steptime .eq. 0) then
        Finn = Fin
        snn  = sn
        call MatProdInv(Fnn,Finn,Fenn)
        call ElasticStress(10*mprop,Fenn,Tnn)
      goto 12
      end if
     
c     We are not in the first increment -> Viscoplastic response  
       call FlowRule(mprop,Fen,
     1     Fin,DFin,sn,Dsn)
           Fi_ave = Fin + 0.5*DT*DFin
           s_ave  = sn  + 0.5*DT*Dsn
           F_ave = 0.5*(Fnn + Fn)    
      call MatProdInv(F_ave,Fi_ave,Fe_ave)
      call FlowRule(mprop,Fe_ave,
     1     Fi_ave,DFi_ave,s_ave,Ds_ave)
           Finn = Fin + DT * DFi_ave  !Finn using Midpoint scheeme. 
           snn  = sn  + DT * Ds_ave
            
C     uncomment "goto 10" to accept the calculations from Midpoint
C     and skip the "accuracy check and iteration scheme". 

         goto 10      
           
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC           
CCCCCCCCCC ACCURACY CHECK AND ITERATION CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
               
           FFE = Fin + dt*DFin      !Finn using Forward Euler(1. order)
      call MatNorm(FFE-Finn,temp1)
      call MatNorm(Finn-Fin,temp2)
      if (temp2 .lt. diff) goto 10
              eps = temp1/temp2  
          if (eps .ne. eps) then
              print*, " "
              print*, "epsilon = NaN"
              print*, "Finn is not defined correctly"
              print*, "possible cause: Flow Rule blows up"
              pause
          end if    
          if (eps .gt. kk) then    !Error is too big!
              dt_new = b*DT*(kk/eps)
              dt_temp = min(dt_new,c*DT)
              iter = ceiling(DT/dt_temp)
              dt_sugg = DT/iter
c             Same theory but with lower timestep
   20             if (iter .gt. MaxIter) goto 10
                  Fn2  = Fn              
                  Fen2 = Fen
                  Fin2 = Fin
                  sn2  = sn
               do k = 1,iter
                          k2 = k
                          iter2 = iter
                          Fnn2 = Fn + (k2/iter2)*(Fnn-Fn)
                          F_ave = 0.5 * (Fnn2 + Fn2)
                     call FlowRule(mprop,Fen2,
     1                    Fin2,DFin2,sn2,Dsn2)
                          Fi_ave = Fin2 + 0.5*dt_sugg * Dfin2
                          s_ave  = sn2  + 0.5*dt_sugg * Dsn2
                     call MatProdInv(F_ave,Fi_ave,Fe_ave)
                     call FlowRule(mprop,Fe_ave,
     1                    Fi_ave,DFi_ave,s_ave,Ds_ave)
                          Finn2 = Fin2 + dt_sugg * DFi_ave
                          snn2 = sn2 + dt_sugg * Ds_ave  
                     call MatProdInv(Fnn2,Finn2,Fenn2)      
                                   !Checking Error again:
                                   FFE = Fin2 + dt_sugg*DFin2
                              call MatNorm(FFE-Finn2,temp1)
                              call MatNorm(Finn2-Fin2,temp2)
                              if (temp2 .lt. MinItStep) goto 30     
                                  eps = temp1/temp2    
                              if (eps .gt. kk) then          
                                  dt_new = b*dt_sugg*(kk/eps)
                                  dt_temp = min(dt_new,c*dt_sugg)
                                  iter = ceiling(DT/dt_temp)
                                  dt_sugg = DT/iter
                                  goto 20
                              end if
  30                      Fn2 = Fnn2
                          Fen2 = Fenn2
                          Fin2 = Finn2
                          sn2  = snn2
              end do 
                 Fenn = Fenn2
                 Finn = Finn2
                 snn  = snn2
          end if   
                 
CCCCCCCCCC END ACCURACY CHECK AND ITERATION CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC           
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC          
          
      !calculate Fenn using Fnn and Finn, and calculate the stress Tnn    
  10  call MatProdInv(Fnn,Finn,Fenn)
      call ElasticStress(mprop,Fenn,Tnn)
      
c     Rotate the stress tensor:      
  12  call TMatProd3(R,Tnn,R,Tnn)
  
C**************************************************************************
C******** Main Calculations Finished -> Storing variables *****************
C**************************************************************************  
  
c     Updating the stress  
      stressnew(km,1) = Tnn(1,1)
      stressnew(km,2) = Tnn(2,2)
      stressnew(km,3) = Tnn(3,3)
      stressnew(km,4) = Tnn(1,2)    
      
c     store statevariables
      do i = 1,3
        Statenew(km,i) = Finn(i,i) - 1.0
      end do
      Statenew(km,4) = Finn(1,2)
      Statenew(km,5) = Finn(2,1)
      StateNew(km,6) = snn - s0

  100 continue
      

      return
      end
 
 
      
C**************************************************************************
C**************** Arruda Boyce Theory - subroutines ***********************
C**************************************************************************


CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
CC            FLOWRULE AND VISCOPLASTIC DAMPENING                          CC
CC            See paper #1, page: 9-10   or(705-706)                       CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
      Subroutine FlowRule(mprop,Fe,Fi,DFi,s,sdot)
          implicit none
          Double precision :: Tr_Tvp,tau,gammadotp
          Double precision :: Tr_Dev_Tvp2,s,sdot,temp,const
          Double precision :: gammadot0,h,flowconst,s0,sss
          Double precision,Dimension(3,3)::F,Fe,Fi,DFi,Te
          Double precision,Dimension(3,3)::B,Be,Bi,Tb
          Double precision,Dimension(3,3)::Tvp, Dev_Tvp, dev_Tvp2
          Double precision,Dimension(3,3)::Identity,Nij,D,R
          Double precision,Dimension(8)::mprop  
          integer::k,l         
c     material parameters                    
          gammadot0         =   mprop(5)
          h                 =   mprop(6)
          sss               =   mprop(7)
          flowconst         =   mprop(8)
c     Preparations   
          Tr_Tvp = 0.0
          temp   = 0.0
          tau    = 0.0
          do k = 1,3
            do l = 1,3
              D(k,l)        = 0.0
              Identity(k,l) = 0.0
              Te(k,l)       = 0.0
              Tvp(k,l)      = 0.0
              dev_Tvp(k,l)  = 0.0
              Nij(k,l)      = 0.0
            end do 
          end do     
C     Main calculations          
          call MatIdent(Identity)
          call ElasticStress(mprop,Fe,Te)
          call FeBackStressFeT(mprop,Fe,Fi,Tb)
               Tvp = Te - Tb  
          call MatTrace(Tvp,Tr_Tvp)         
               dev_Tvp = Tvp - (Tr_Tvp/3)*Identity
               do k = 1,3
                do l = 1,3
                   temp = temp + dev_tvp(k,l)**2.0 
                end do 
               end do
               tau = sqrt(0.5)*sqrt(temp)   
          if (tau .gt. 0) then      !We have plastic deformation
                Nij=(sqrt(0.5)/tau)*dev_Tvp 
          end if 
               gammadotp=gammadot0*
     1         exp(-(flowconst*s)*(1-((tau/s)**(5.0/6.0))))
               D = gammadotp*Nij
          call MatProd(D,Fi,DFi)
               sdot = h*(1-(s/sss))*gammadotp
          return 
      end subroutine FlowRule


CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
CC              CALCULATE THE BACKSTRESS                                   CC
CC              See paper #2, page: 19-20   or(11-21)                      CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
      Subroutine FeBackStressFeT(mprop,Fe,Fi,Tb)
      implicit none 
          Double precision::Tr_Bi, Lambdapch, Cr, N, temp, IL,I,IB   
          Double precision,Dimension(3,3)::Fe,Fi,Fe_Tb_FeT,Bi
          Double precision,Dimension(8)::mprop
          Double precision,dimension(3,3)::Q,BB,Tb
          Double precision,dimension(3)::EigBi
          Double precision::third,detFe
          integer::k,l
c     material parameters          
          Cr           = mprop(3)
          N            = mprop(4)
c     Preparations
          Lambdapch = 0.0
          temp      = 0.0
          IL        = 0.0 
          do k = 1,3
                eigBi(k) = 0.0
                do l = 1,3
                    BB(k,l)        = 0.0
                    Bi(k,l)        = 0.0
                    Q(k,l)         = 0.0
                    Fe_Tb_FeT(k,l) = 0.0
                end do 
          end do
C     Main Calculations          
          call MatDet(Fe,detFe)         
          call MatProdT(Fi,Fi,Bi)
          call MatTrace(Bi,Tr_Bi)
          call DSYEVQ3(Bi,Q,eigBi)
               IB = eigBi(1) + eigBi(2) + eigBi(3)  
               Lambdapch=sqrt(IB/3)
               temp=Lambdapch/sqrt(N)
               call InvLang(Temp,IL)
            do k=1,3
         BB(k,k)=Cr/3.0*sqrt(N)*IL*(eigBi(k)-IB/3)/Lambdapch
            end do 
          call MatProdT3(Fe,BB,Fe,Fe_Tb_FeT)
          Tb = (1.0/detFe)*Fe_Tb_FeT
          return 
      end subroutine FeBackStressFeT                    
             
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
CC          CALCULATE THE ELASTIC STRESS RESPONSE                          CC
CC      This subroutine uses the material properties and the elastic       CC
CC      deformation gradient to calculate the elastic stress               CC
CC      See paper #1, page: 8     or(704)                                  CC 
CC      See paper #3, page: 2-3   or(52-53)                                CC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC 
      Subroutine ElasticStress(mprop,Fe,Te)
      include 'VABA_PARAM.INC'
          Double precision :: DetFe,G,Lambda,Tr_lnsqrtBe       
          Double precision,Dimension(3,3)::Te,Fe,Be,lnsqrtBe
          Double precision,Dimension(3,3)::temp
          Double precision,Dimension(8)::mprop
          integer :: i,j           
c     Material properties:
          G       = mprop(1)
          Lambda  = mprop(2)      
c     preparations: 
          DetFe = 0.0              
          do i = 1,3
            do j = 1,3
                Te(i,j)       = 0.0
                temp(i,j)     = 0.0
                Be(i,j)       = 0.0
                lnsqrtBe(i,j) = 0.0
            end do
          end do           
c     Main calculations                       
          call MatDet(Fe,DetFe)
          call MatProdT(Fe,Fe,Be)
          !Using Pade approximation to find: henky strain = ln(sqrt(Be))
          !call PadeApprox(Be,lnsqrtBe)
          call PadeExact(Be,lnsqrtBe)
          tr_lnsqrtBe = lnsqrtBe(1,1)+lnsqrtBe(2,2)+lnsqrtBe(3,3)
          do i = 1,3
            do j = 1,3
                Te(i,j) = 2*G*lnsqrtBe(i,j)
                if (i .eq. j) then
                    Te(i,j) = Te(i,j) + Lambda*tr_lnsqrtBe
                end if
            end do 
          end do  
          temp = Te
          Te = (1.0/DetFe)*Temp 
          return 
      end subroutine ElasticStress  

C**************************************************************************
C********************* Numerical approximations ***************************
C**************************************************************************

C     APPROXIMATE (ln(sqrt(Mat)))
c       This subroutine uses Pade approximation
c       to find (ln(sqrt(Mat))) of Mat      
      subroutine PadeApprox(Mat,lnsqrtMat)
      include 'VABA_PARAM.INC'
      double precision,dimension(3,3)::I,lnsqrtMat,Mat
      double precision,dimension(3,3)::a,b,c,c_inv,Mat2,y,x
      call MatIdent(I)
      a=Mat+I
      b=Mat-I
      call MatProd(Mat,Mat,Mat2)
      c = 2 * ( Mat2 + 4 * Mat + I)
      call MatInv(c,c_inv)
      call MatProd(a,b,x)
      call MatProd(x,c_inv,y)
      lnsqrtMat=3*y
      end subroutine PadeApprox
      
C     APPROXIMATE (ln(sqrt(Mat)))
c       This subroutine uses Pade approximation
c       to find (ln(sqrt(Mat))) of Mat      
      subroutine PadeExact(Mat,lnsqrtMat)
      include 'VABA_PARAM.INC'
      double precision,dimension(3,3)::Mat,lnsqrtMat,Q,temp
      double precision,dimension(3)::W,Q_inv
      call DSYEVQ3(Mat,Q,W)
      call MatIdent(temp)
      temp(1,1) = 0.5*log(W(1))
      temp(2,2) = 0.5*log(W(2))
      temp(3,3) = 0.5*log(W(3))
      Call MatProdInv3(Q,temp,Q,lnsqrtMat)
      end subroutine PadeExact      
      
C     This gives an approximation to the inverse langevin function
C     The function takes in: a, and gives the result as b
C     within 5% accuracy 
      Subroutine InvLang(x,L)
      include 'VABA_PARAM.INC'
        double precision :: x,L,a,b,c
        a = 2.99248834685337 
        b = -1.14365108190676
        c = -1.000
        L = x*((a+b*x*x)/(1+c*x*x))        
        return
      end Subroutine InvLang 
      
     
C**************************************************************************
C****************** SUBROUTINES - Matrix operations ***********************
C**************************************************************************            

C     calculate the product of a*b*trans(c):       
      Subroutine MatProdInv3(a,b,c,d)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,d,x,y
      call MatProd(a,b,x)
      call MatInv(c,y)
      call MatProd(x,y,d)
      return 
      end subroutine MatProdInv3

C     calculate the product of trans(a)*b*c:       
      Subroutine TMatProd3(a,b,c,d)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,d,x,y
      call MatTrans(a,y)
      call MatProd(y,b,x)
      call MatProd(x,c,d)
      return 
      end subroutine TMatProd3
      
C     calculate the product of a*b*trans(c):       
      Subroutine MatProdT3(a,b,c,d)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,d,x,y
      call MatProd(a,b,x)
      call MatTrans(c,y)
      call MatProd(x,y,d)
      return 
      end subroutine MatProdT3      
      
C     calculate the product of a matrix a and inv(b)     
      Subroutine MatProdInv(a,b,c)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,x
      call MatInv(b,x)
      call MatProd(a,x,c)
      return 
      end subroutine MatProdInv         

C     calculate the product of a matrix a and trans(b)     
      Subroutine MatProdT(a,b,c)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c,x
      call MatTrans(b,x)
      call MatProd(a,x,c)
      return 
      end subroutine MatProdT 

C     Calculating the inverse of a 3x3 matrix
      Subroutine MatInv(a,b)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3) :: a,b,x,z 
      Double precision :: y
      call MatDet(a,y)
      call MatTrans(a,x)
          z(1,1) =  +(x(2,2)*x(3,3) - x(2,3)*x(3,2))
          z(1,2) =  -(x(2,1)*x(3,3) - x(2,3)*x(3,1))
          z(1,3) =  +(x(2,1)*x(3,2) - x(2,2)*x(3,1))
          z(2,1) =  -(x(1,2)*x(3,3) - x(1,3)*x(3,2))
          z(2,2) =  +(x(1,1)*x(3,3) - x(1,3)*x(3,1))
          z(2,3) =  -(x(1,1)*x(3,2) - x(1,2)*x(3,1))
          z(3,1) =  +(x(1,2)*x(2,3) - x(1,3)*x(2,2))
          z(3,2) =  -(x(1,1)*x(2,3) - x(1,3)*x(2,1))
          z(3,3) =  +(x(1,1)*x(2,2) - x(1,2)*x(2,1))
      b = (1/y) * z    
      return
      end subroutine MatInv  
      
C     calculating transpose of matrix
      Subroutine MatTrans(a,b)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3) :: a,b
      integer j,k
      do j=1,3
        do k=1,3
            b(j,k) = a(k,j)
        end do 
      end do
      return
      end Subroutine MatTrans      
      
C     Calculate the determinant of a matrix
      Subroutine MatDet(a,det)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a 
      Double precision :: det
      det = a(1,1)*a(2,2)*a(3,3)-a(1,1)*a(2,3)*a(3,2)-
     1 a(1,2)*a(2,1)*a(3,3)+ a(1,2)*a(2,3)*a(3,1)+
     2 a(1,3)*a(2,1)*a(3,2)- a(1,3)*a(2,2)*a(3,1)
      return
      end Subroutine MatDet
      
C     Calculating product of two 3x3 matrices:
      Subroutine MatProd(a,b,c)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::a,b,c
          c(1,1) = a(1,1)*b(1,1) + a(1,2)*b(2,1) + a(1,3)*b(3,1)
          c(1,2) = a(1,1)*b(1,2) + a(1,2)*b(2,2) + a(1,3)*b(3,2)
          c(1,3) = a(1,1)*b(1,3) + a(1,2)*b(2,3) + a(1,3)*b(3,3)
          c(2,1) = a(2,1)*b(1,1) + a(2,2)*b(2,1) + a(2,3)*b(3,1)
          c(2,2) = a(2,1)*b(1,2) + a(2,2)*b(2,2) + a(2,3)*b(3,2)
          c(2,3) = a(2,1)*b(1,3) + a(2,2)*b(2,3) + a(2,3)*b(3,3)
          c(3,1) = a(3,1)*b(1,1) + a(3,2)*b(2,1) + a(3,3)*b(3,1)
          c(3,2) = a(3,1)*b(1,2) + a(3,2)*b(2,2) + a(3,3)*b(3,2)
          c(3,3) = a(3,1)*b(1,3) + a(3,2)*b(2,3) + a(3,3)*b(3,3)
      return
      end Subroutine MatProd
      
c     This makes a 3x3 identity matrix      
      Subroutine MatIdent(I)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)::I
      integer::k,l
      do k = 1,3
         do l = 1,3
            if (k .eq. l) then
                I(k,l) = 1.0
            else
                I(k,l) = 0.0
            end if
         end do 
      end do 
      return
      end Subroutine MatIdent
      
c     This subroutine calculates the trace of a matrix
      Subroutine MatTrace(Mat,Trace)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)   ::  Mat
      Double precision                  ::  Trace
      Trace = Mat(1,1)+Mat(2,2)+Mat(3,3)
      return
      end Subroutine MatTrace
      
c     This subroutine calculates the norm of a matrix
      Subroutine MatNorm(Mat,Norm)
      include 'VABA_PARAM.INC'
      Double precision,Dimension(3,3)   ::  Mat
      Double precision                  ::  temp,Norm
      integer::i,j
      temp = 0
      do i = 1,3
        do j = 1,3
            temp = temp + Mat(i,j)**2
        end do 
      end do
      Norm = sqrt(temp)
      return
      end Subroutine MatNorm      
            
      
C**************************************************************************
C****************  EIGENVALUES and EIGENMATRIX ****************************
C**************************************************************************            


      SUBROUTINE DSYEVQ3(A, Q, W)
      include 'VABA_PARAM.INC'
* ----------------------------------------------------------------------------
* Calculates the eigenvalues and normalized eigenvectors of a symmetric 3x3
* matrix A using the QL algorithm with implicit shifts, preceded by a
* Householder reduction to real tridiagonal form.
* The function accesses only the diagonal and upper triangular parts of
* A. The access is read-only. 
* ----------------------------------------------------------------------------
* Parameters:
*   A: The symmetric input matrix
*   Q: Storage buffer for eigenvectors
*   W: Storage buffer for eigenvalues
* ----------------------------------------------------------------------------
* Dependencies:
*   DSYTRD3()
* ----------------------------------------------------------------------------
*     .. Arguments ..
      DOUBLE PRECISION A(3,3)
      DOUBLE PRECISION Q(3,3)
      DOUBLE PRECISION W(3)

*     .. Parameters ..
      INTEGER          N
      PARAMETER        ( N = 3 )

*     .. Local Variables ..
      DOUBLE PRECISION E(3)
      DOUBLE PRECISION G, R, P, F, B, S, C, T
      INTEGER          NITER
      INTEGER          L, M, I, J, K

*     .. External Functions ..
      EXTERNAL         DSYTRD3
      
*     Transform A to real tridiagonal form by the Householder method
      CALL DSYTRD3(A, Q, W, E)

*     Calculate eigensystem of the remaining real symmetric tridiagonal
*     matrix with the QL method
*
*     Loop over all off-diagonal elements
      DO 10 L = 1, N-1
        NITER = 0

*       Iteration loop
        DO 11 I = 1, 50
*         Check for convergence and exit iteration loop if off-diagonal
*         element E(L) is zero
          DO 20 M = L, N-1
            G = ABS(W(M)) + ABS(W(M+1))
            IF (ABS(E(M)) + G .EQ. G) THEN
              GO TO 30
            END IF
   20     CONTINUE
   30     IF (M .EQ. L) THEN
            GO TO 10
          END IF

          NITER = NITER + 1
          IF (NITER >= 100) THEN
            PRINT *, 'DSYEVQ3: No convergence.'
            RETURN
          END IF

*         Calculate G = D(M) - K
          G = (W(L+1) - W(L)) / (2.0D0 * E(L))
          R = SQRT(1.0D0 + G**2)
          IF (G .GE. 0.0D0) THEN
            G = W(M) - W(L) + E(L)/(G + R)
          ELSE
            G = W(M) - W(L) + E(L)/(G - R)
          END IF

          S = 1.0D0
          C = 1.0D0
          P = 0.0D0
          DO 40 J = M - 1, L, -1
            F = S * E(J)
            B = C * E(J)
            IF (ABS(F) .GT. ABS(G)) THEN
              C      = G / F
              R      = SQRT(1.0D0 + C**2)
              E(J+1) = F * R
              S      = 1.0D0 / R
              C      = C * S
            ELSE
              S      = F / G
              R      = SQRT(1.0D0 + S**2)
              E(J+1) = G * R
              C      = 1.0D0 / R
              S      = S * C
            END IF

            G      = W(J+1) - P
            R      = (W(J) - G) * S + 2.0D0 * C * B
            P      = S * R
            W(J+1) = G + P
            G      = C * R - B

*           Form eigenvectors
*           --- This loop can be omitted if only the eigenvalues are desired ---
            DO 50 K = 1, N
              T         = Q(K, J+1)
              Q(K, J+1) = S * Q(K, J) + C * T
              Q(K, J)   = C * Q(K, J) - S * T
   50       CONTINUE
   40     CONTINUE
          W(L) = W(L) - P
          E(L) = G
          E(M) = 0.0D0
   11   CONTINUE
   10 CONTINUE
  
      END SUBROUTINE

* ----------------------------------------------------------------------------
      SUBROUTINE DSYTRD3(A, Q, D, E)
      include 'VABA_PARAM.INC'
* ----------------------------------------------------------------------------
* Reduces a symmetric 3x3 matrix to real tridiagonal form by applying
* (unitary) Householder transformations:
*            [ D[1]  E[1]       ]
*    A = Q . [ E[1]  D[2]  E[2] ] . Q^T
*            [       E[2]  D[3] ]
* The function accesses only the diagonal and upper triangular parts of
* A. The access is read-only.
* ---------------------------------------------------------------------------
*     .. Arguments ..
      DOUBLE PRECISION A(3,3)
      DOUBLE PRECISION Q(3,3)
      DOUBLE PRECISION D(3)
      DOUBLE PRECISION E(2)

*     .. Parameters ..
      INTEGER          N
      PARAMETER        ( N = 3 )

*     .. Local Variables ..
      DOUBLE PRECISION U(N), P(N)
      DOUBLE PRECISION OMEGA, F
      DOUBLE PRECISION K, H, G
      INTEGER          I, J

*     Initialize Q to the identitity matrix
*     --- This loop can be omitted if only the eigenvalues are desired ---
      DO 10 I = 1, N
        Q(I,I) = 1.0D0
        DO 11, J = 1, I-1
          Q(I, J) = 0.0D0
          Q(J, I) = 0.0D0
   11   CONTINUE
   10 CONTINUE

*     Bring first row and column to the desired form
      H = A(1,2)**2 + A(1,3)**2
      IF (A(1,2) .GT. 0.0D0) THEN
        G = -SQRT(H)
      ELSE
        G = SQRT(H)
      END IF
      E(1)  = G
      F     = G * A(1,2)
      U(2)  = A(1,2) - G
      U(3)  = A(1,3)

      OMEGA = H - F
      IF (OMEGA > 0.0D0) THEN
        OMEGA = 1.0D0 / OMEGA
        K     = 0.0D0
        DO 20 I = 2, N
          F    = A(2,I)*U(2) + A(I,3)*U(3)
          P(I) = OMEGA * F
          K    = K + U(I) * F
  20    CONTINUE
        K = 0.5D0 * K * OMEGA**2

        DO 30 I = 2, N
          P(I) = P(I) - K * U(I)
  30    CONTINUE

        D(1) = A(1,1)
        D(2) = A(2,2) - 2.0D0 * P(2) * U(2)
        D(3) = A(3,3) - 2.0D0 * P(3) * U(3)

*       Store inverse Householder transformation in Q
*       --- This loop can be omitted if only the eigenvalues are desired ---
        DO 40, J = 2, N
          F = OMEGA * U(J)
          DO 41 I = 2, N
            Q(I,J) = Q(I,J) - F * U(I)
   41     CONTINUE
   40   CONTINUE
            
*       Calculated updated A(2, 3) and store it in E(2)
        E(2) = A(2, 3) - P(2) * U(3) - U(2) * P(3)
      ELSE
        DO 50 I = 1, N
          D(I) = A(I, I)
  50    CONTINUE
        E(2) = A(2, 3)
      END IF
      
      END SUBROUTINE
* End of subroutine DSYTRD3

